#include "MeshLoader.h"
#include <iomanip>
#include <algorithm>
#include <set>
#include <unordered_set>

std::vector<FiniteElement> MeshLoader::finite_elements_ID(int ID1, int ID2, int ID3) {
    std::vector<FiniteElement> result;
    auto iter = finite_elements.begin();
    auto last_iter = finite_elements.end();

    while (iter != last_iter) {
        iter = std::find_if(iter, last_iter, 
            [ID1, ID2, ID3](FiniteElement& elem) { //!!! Где const 
            return std::find(elem.nodes_ID.begin(), elem.nodes_ID.end(), ID1) != elem.nodes_ID.end()
                || std::find(elem.nodes_ID.begin(), elem.nodes_ID.end(), ID2) != elem.nodes_ID.end()
                || std::find(elem.nodes_ID.begin(), elem.nodes_ID.end(), ID3) != elem.nodes_ID.end();
            });

        if (iter != last_iter) {
            result.push_back(*iter);
            ++iter;
        }
    }

    return result;
}

std::vector<FiniteElement> MeshLoader::finite_elements_Edge(int ID1, int ID2) {
    std::vector<FiniteElement> result;
    auto iter = finite_elements.begin();
    auto last_iter = finite_elements.end();

    while (iter != last_iter) {
        iter = std::find_if(iter, last_iter, 
        [ID1, ID2](FiniteElement& element) { //!!! Где const 
            return std::find(element.nodes_ID.begin(), element.nodes_ID.end(), ID1) != element.nodes_ID.end()
                || std::find(element.nodes_ID.begin(), element.nodes_ID.end(), ID2) != element.nodes_ID.end();
        });

        if (iter != last_iter) {
            result.push_back(*iter);
            ++iter;
        }
    }

    return result;
}

//!!! Все контейнера должны содержать только ID узлов или КЭ (Не исправлено)

std::vector<Node> MeshLoader::vertex_nodes_by_surface_ID(int ID) {
    std::set<Node> result;
    auto iter = surfaces.begin();
    auto last_iter = surfaces.end();

    while (iter != last_iter) {
        iter = std::find_if(iter, last_iter, 
        [ID](const BoundaryFiniteElement& elem) { 
            return elem.edge_ID == ID;
        });

        if (iter != last_iter) {
            for (const auto& it : iter->nodes_ID) { //!!! it - не итератор. Переименуйте 
                
                //!!! Крайне неэффективная реализация. Потеряны преимущества контейнера (не исправлено)
                auto node_it = std::find_if(nodes.begin(), nodes.end(), [it](const Node& N) { 
                    return N.ID == it;
                });
                result.insert(*node_it);
            }
            ++iter;
        }
    }

    return std::vector<Node>(result.begin(), result.end());
}

std::vector<BoundaryFiniteElement> MeshLoader::surfaces_ID(int ID) {
    std::vector<BoundaryFiniteElement> result; 
    auto iter = surfaces.begin();
    auto last_iter = surfaces.end();

    while (iter != last_iter) {
        iter = std::find_if(iter, last_iter, 
        [ID](const BoundaryFiniteElement& elem) {
            return elem.element_ID == ID;
        });

        if (iter != last_iter) {
            result.push_back(*iter);
            ++iter;
        }
    }

    return std::vector<BoundaryFiniteElement>(result.begin(), result.end());
}

std::vector<FiniteElement> MeshLoader::finite_elements_material(int material_ID) { 
    std::vector<FiniteElement> result;
    auto iter = finite_elements.begin();
    auto last_iter = finite_elements.end();

    while (iter != last_iter) {
        iter = std::find_if(iter, last_iter, 
        [material_ID](const FiniteElement& element) { 
            return element.material_ID == material_ID;
        });

        if (iter != last_iter) {
            result.push_back(*iter);
            ++iter;
        }
    }

    return std::vector<FiniteElement>(result.begin(), result.end());
}

void MeshLoader::middle_insert() {

    std::unordered_set<Edge, Hash> edges;
    for (auto& iter : finite_elements) { //!!! iter не итератор - переименовать
        std::vector<int> N = std::move(iter.nodes_ID); //!!! Копирование зачем? (не исправлено) Только хуже
                                                       //!!! Почему с iter работать нельзя?

        for (auto i = N.begin(); i < N.end(); ++i) { //!!! Лучше использовать индексы

            for (auto j = i + 1; j < N.end(); ++j) { //!!! Лучше использовать индексы
                Edge E(*i, *j, 0);
                
                auto E2 = edges.find(E); 

                if (E2 == edges.end()) {
                    Node left_node = get_node(*i);
                    Node right_node = get_node(*j);
                    int center_id_ = last_nodes() + 1;

                    Node center_node(center_id_,
                        std::array<double, 3>{(left_node.cd[0] + right_node.cd[0]) / 2,
                        (left_node.cd[1] + right_node.cd[1]) / 2,
                        (left_node.cd[2] + right_node.cd[2]) / 2},
                        false);

                    E.UpdateCenter(center_id_);
                    nodes.push_back(std::move(center_node));
                    iter.nodes_ID.push_back(center_id_);
                    edges.insert(E);

                } else {
                    iter.nodes_ID.push_back(E2->center);
                }
            }
        }

        iter.nodes_ID.shrink_to_fit();
    }

    for (auto& it : surfaces) {
        std::vector<int> N(it.nodes_ID);

        for (auto i = N.begin(); i < N.end() - 1; ++i) {
            for (auto j = i + 1; j < N.end(); ++j) {
                Edge E(*i, *j, 0);
                auto E2 = edges.find(E); 

                if (E2 != edges.end()) {
                    it.nodes_ID.push_back(E2->center);
                }
            }
        }

        it.nodes_ID.shrink_to_fit();
    }

    surfaces.shrink_to_fit();
    nodes.shrink_to_fit();
}

std::vector<std::vector<Node>> MeshLoader::near_nodes() {
    std::vector<std::set<int>> res(nodes.size() + 1);

    for (const auto& it : finite_elements) {
        for (auto& iter : it.nodes_ID) {
            for (auto& iter2 : it.nodes_ID) {
                res[iter].insert(iter2);
            }
            res[iter].erase(iter);
        }
    }

    std::vector<std::vector<Node>> result(nodes.size() + 1);
    for (int i = 1; i < nodes.size() + 1; ++i) {
        std::vector<Node> temp;
        temp.reserve(res[i].size());

        for (auto& it : res[i]) {
            temp.push_back(get_node(it));
        }

        result[i] = std::move(temp);
    }

    return result;
}

double MeshLoader::volume(int ID1, int ID2, int ID3, int ID4) {
    Node A = get_node(ID1);
    Node B = get_node(ID2);
    Node C = get_node(ID3);
    Node D = get_node(ID4);
    double A1 = (B.cd[0] - A.cd[0]) * ((C.cd[1] - A.cd[1]) * (D.cd[2] - A.cd[2]) - (C.cd[2] - A.cd[2]) * (D.cd[1] - A.cd[1]));
    double A2 = -(B.cd[1] - A.cd[1]) * ((C.cd[0] - A.cd[0]) * (D.cd[2] - A.cd[2]) - (D.cd[0] - A.cd[0]) * (C.cd[2] - A.cd[2]));
    double A3 = (B.cd[2] - A.cd[2]) * ((C.cd[0] - A.cd[0]) * (D.cd[1] - A.cd[1]) - (D.cd[0] - A.cd[0]) * (C.cd[1] - A.cd[1]));
    return (A1 + A2 + A3) / 6.0;
}

void MeshLoader::volume_range(int m_ID, double a, double b) { 
    
    
    
    //!!! И где тут std::ranges и views ???
    
    
    
    
    std::vector<FiniteElement> elements = finite_elements_material(m_ID);
    if (elements.empty())
        return;
    std::cout << std::endl << "------------------------- Nodes_ID -------------------------" << std::endl;
    for (auto i = elements.begin(); i + 3 != elements.end(); ++i) {
        double V = volume(i->nodes_ID[0], (i + 1)->nodes_ID[0], (i + 2)->nodes_ID[0], (i + 3)->nodes_ID[0]);
        if (V - a > 0.00001 && b - V > 0.00001)
            std::cout << i->nodes_ID[0] << " " << (i + 1)->nodes_ID[0] << " " << (i + 2)->nodes_ID[0] << " " << (i + 3)->nodes_ID[0] << std::endl;
    }
}

